==> divachain-0.27.2.tgz: package/src/chain/blockchain.ts <==
/**
 * Copyright (C) 2021 diva.exchange
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Author/Maintainer: Konrad Bächler <konrad@diva.exchange>
 */

import { BlockStruct } from './block';
import fs from 'fs';
import LevelUp from 'levelup';
import LevelDown from 'leveldown';
import path from 'path';
import {
  CommandAddPeer,
  CommandRemovePeer,
  CommandModifyStake,
  CommandData,
  CommandDecision,
  TransactionStruct,
} from './transaction';
import { Server } from '../net/server';
import { Logger } from '../logger';
import { Util } from './util';

export type Peer = {
  publicKey: string;
  http: string;
  udp: string;
  stake: number;
};

export class Blockchain {
  public static readonly COMMAND_ADD_PEER = 'addPeer';
  public static readonly COMMAND_REMOVE_PEER = 'removePeer';
  public static readonly COMMAND_MODIFY_STAKE = 'modifyStake';
  public static readonly COMMAND_DATA = 'data';
  public static readonly COMMAND_DECISION = 'decision';
  public static readonly STATE_DECISION_IDENT = 'decision:';
  public static readonly STATE_PEER_IDENT = 'peer:';
  public static readonly STATE_DECISION_TAKEN = 'decision:taken:';

  private readonly server: Server;
  private readonly publicKey: string;
  private readonly dbBlockchain: InstanceType<typeof LevelUp>;
  private readonly dbState: InstanceType<typeof LevelUp>;

  private height: number = 0;
  private mapBlocks: Map<number, BlockStruct> = new Map();
  private latestBlock: BlockStruct = {} as BlockStruct;

  private mapPeer: Map<string, Peer> = new Map();

  private quorum: number = 0;

  static async make(server: Server): Promise<Blockchain> {
    const b = new Blockchain(server);
    if (server.config.bootstrap) {
      await b.clear();
    } else {
      await b.init();
    }
    return b;
  }

  private constructor(server: Server) {
    this.server = server;
    this.publicKey = this.server.getWallet().getPublicKey();

    this.dbBlockchain = LevelUp(LevelDown(path.join(this.server.config.path_blockstore, this.publicKey)), {
      createIfMissing: true,
      errorIfExists: false,
      compression: true,
      cacheSize: 2 * 1024 * 1024, // 2 MB
    });

    this.dbState = LevelUp(LevelDown(path.join(this.server.config.path_state, this.publicKey)), {
      createIfMissing: true,
      errorIfExists: false,
      compression: true,
      cacheSize: 2 * 1024 * 1024, // 2 MB
    });
  }

  private async init() {
    this.height = 0;
    this.mapBlocks = new Map();
    this.latestBlock = {} as BlockStruct;
    await this.dbState.clear();

    await new Promise((resolve, reject) => {
      this.dbBlockchain
        .createReadStream()
        .on('data', async (data) => {
          const block: BlockStruct = JSON.parse(data.value) as BlockStruct;
          this.updateCache(block);
          await this.processState(block);
        })
        .on('end', resolve)
        .on('error', reject);
    });
  }

  async shutdown() {
    await this.dbBlockchain.close();
    await this.dbState.close();
  }

  async clear() {
    await this.dbBlockchain.clear();
    await this.dbState.clear();

    this.height = 0;
    this.mapBlocks = new Map();
    this.latestBlock = {} as BlockStruct;
    this.mapPeer = new Map();
  }

  async reset(genesis: BlockStruct) {
    await this.clear();
    await this.updateBlockData(String(1).padStart(16, '0'), JSON.stringify(genesis));
    await this.init();
  }

  add(block: BlockStruct): boolean {
    if (
      this.height + 1 !== block.height ||
      block.previousHash !== this.latestBlock.hash ||
      !this.server.getValidation().validateBlock(block)
    ) {
      return false;
    }

    this.updateCache(block);

    (async (b: BlockStruct) => {
      await this.updateBlockData(String(b.height).padStart(16, '0'), JSON.stringify(b));
      await this.processState(b);
    })(block);

    return true;
  }

  private updateCache(block: BlockStruct) {
    this.height = block.height;
    this.latestBlock = block;

    // cache
    this.mapBlocks.set(this.height, block);
    if (this.mapBlocks.size > this.server.config.blockchain_max_blocks_in_memory) {
      this.mapBlocks.delete(this.height - this.server.config.blockchain_max_blocks_in_memory);
    }
  }

  async getRange(gte: number, lte: number): Promise<Array<BlockStruct>> {
    lte = Math.floor(lte < 1 ? this.height : lte);
    gte = gte <= this.height ? (gte < 1 ? 1 : Math.floor(gte)) : this.height;
    lte = lte <= this.height ? lte : this.height;
    gte = lte - gte > 0 ? gte : lte;
    gte = lte - gte >= this.server.config.api_max_query_size ? lte - this.server.config.api_max_query_size + 1 : gte;

    // cache available?
    if (this.mapBlocks.has(gte) && this.mapBlocks.has(lte)) {
      const start = this.mapBlocks.size - this.height + gte - 1;
      const end = start + lte - gte + 1;
      return [...this.mapBlocks.values()].slice(start, end);
    }

    const a: Array<BlockStruct> = [];
    return new Promise((resolve, reject) => {
      this.dbBlockchain
        .createValueStream({ gte: String(gte).padStart(16, '0'), lte: String(lte).padStart(16, '0') })
        .on('data', (data) => {
          a.push(JSON.parse(data));
        })
        .on('end', () => {
          resolve(a);
        })
        .on('error', reject);
    });
  }

  async getPage(page: number, size: number, filter: RegExp | false = false): Promise<Array<BlockStruct>> {
    page = page < 1 ? 1 : Math.floor(page);
    size =
      size < 1 || size > this.server.config.api_max_query_size
        ? this.server.config.api_max_query_size
        : Math.floor(size);

    const aFiltered = filter ? await this.filter(filter) : [];
    const height = aFiltered.length || this.height;

    let gte = height - page * size + 1;
    gte = gte < 1 ? 1 : gte;

    return filter ? Promise.resolve(aFiltered.slice(gte - 1, gte + size - 1)) : this.getRange(gte, gte + size - 1);
  }

  private async filter(f: RegExp): Promise<Array<BlockStruct>> {
    return new Promise((resolve, reject) => {
      const a: Array<BlockStruct> = [];
      this.dbBlockchain
        .createValueStream()
        .on('data', (data) => {
          const b = JSON.parse(data);
          f.test(JSON.stringify(b)) && a.push(b);
        })
        .on('end', () => {
          resolve(a);
        })
        .on('error', reject);
    });
  }

  async getTransaction(origin: string, ident: string): Promise<{ height: number; transaction: TransactionStruct }> {
    return new Promise((resolve, reject) => {
      // cache
      for (const b of [...this.mapBlocks.values()]) {
        const t = b.tx.find((t: TransactionStruct) => t.origin === origin && t.ident === ident);
        if (t) {
          return resolve({ height: b.height, transaction: t });
        }
      }

      // disk
      let b: BlockStruct = {} as BlockStruct;
      let t: TransactionStruct = {} as TransactionStruct;
      this.dbBlockchain
        .createValueStream()
        .on('data', (data) => {
          if (!t.origin) {
            b = JSON.parse(data) as BlockStruct;
            t =
              b.tx.find((t: TransactionStruct) => t.origin === origin && t.ident === ident) ||
              ({} as TransactionStruct);
          }
        })
        .on('end', () => {
          t.origin ? resolve({ height: b.height, transaction: t }) : reject(new Error('Not Found'));
        })
        .on('error', reject);
    });
  }

  async getState(key: string): Promise<Array<{ key: string; value: any }>> {
    return new Promise((resolve) => {
      if (!key.length) {
        const a: Array<any> = [];
        this.dbState
          .createReadStream()
          .on('data', (data) => {
            a.push({
              key: data.key.toString(),
              value: data.value.toString(),
            });
            if (a.length === this.server.config.api_max_query_size) {
              return resolve(a);
            }
          })
          .on('end', () => {
            resolve(a);
          })
          .on('error', () => {
            resolve([]);
          });
      } else {
        this.dbState.get(key, (error, value: Buffer) => {
          error ? resolve([]) : resolve([{ key: key, value: value.toString() }]);
        });
      }
    });
  }

  getLatestBlock(): BlockStruct {
    return this.latestBlock;
  }

  getHeight(): number {
    return this.height;
  }

  getStake(publicKey: string): number {
    try {
      return this.getPeer(publicKey).stake;
    } catch (error) {
      return 0;
    }
  }

  getTotalQuorum(): number {
    if (this.quorum <= 0) {
      throw new Error('Invalid network quorum');
    }

    return this.quorum;
  }

  getQuorum(): number {
    if (this.quorum <= 0) {
      throw new Error('Invalid network quorum');
    }

    return (2 * this.quorum) / 3; // PBFT, PoS
  }

  getMapPeer(): Map<string, Peer> {
    return this.mapPeer;
  }

  hasPeer(publicKey: string): boolean {
    return this.mapPeer.has(publicKey);
  }

  getPeer(publicKey: string): Peer {
    return this.mapPeer.get(publicKey) as Peer;
  }

  hasNetworkHttp(http: string): boolean {
    for (const v of [...this.mapPeer]) {
      if (v[1].http === http) {
        return true;
      }
    }
    return false;
  }

  async getPerformance(height: number): Promise<{ timestamp: number }> {
    let ts: number;
    try {
      ts = Number((await this.dbState.get('debug-performance-' + height)).toString());
    } catch (error) {
      ts = 0;
    }
    return { timestamp: ts };
  }

  static genesis(p: string): BlockStruct {
    if (!fs.existsSync(p)) {
      throw new Error('Genesis Block not found at: ' + p);
    }
    const b: BlockStruct = JSON.parse(fs.readFileSync(p).toString());
    b.hash = Util.hash(b.previousHash + b.version + b.height + JSON.stringify(b.tx));
    return b;
  }

  private async processState(block: BlockStruct) {
    if (this.server.config.debug_performance) {
      await this.updateStateData('debug-performance-' + this.height, new Date().getTime().toString());
    }

    for (const t of block.tx) {
      for (const c of t.commands) {
        switch (c.command) {
          case Blockchain.COMMAND_ADD_PEER:
            await this.addPeer(c as CommandAddPeer);
            break;
          case Blockchain.COMMAND_REMOVE_PEER:
            await this.removePeer(c as CommandRemovePeer);
            break;
          case Blockchain.COMMAND_MODIFY_STAKE:
            await this.modifyStake(c as CommandModifyStake);
            break;
          case Blockchain.COMMAND_DATA:
            await this.updateStateData((c as CommandData).ns + ':' + t.origin, (c as CommandData).d);
            break;
          case Blockchain.COMMAND_DECISION:
            await this.setDecision((c as CommandDecision).ns, t.origin, (c as CommandDecision).d);
            break;
        }
      }
    }
  }

  private async addPeer(command: CommandAddPeer) {
    if (this.mapPeer.has(command.publicKey)) {
      return;
    }

    const peer: Peer = {
      publicKey: command.publicKey,
      http: command.http,
      udp: command.udp,
      stake: 0,
    };
    this.mapPeer.set(command.publicKey, peer);
    await this.updateStateData(Blockchain.STATE_PEER_IDENT + command.publicKey, peer.stake.toString());
  }

  private async removePeer(command: CommandRemovePeer) {
    if (this.mapPeer.has(command.publicKey)) {
      const peer: Peer = this.mapPeer.get(command.publicKey) as Peer;
      this.quorum = this.quorum - peer.stake;

      this.mapPeer.delete(command.publicKey);
      await this.dbState.del(Blockchain.STATE_PEER_IDENT + command.publicKey);
    }
  }

  private async modifyStake(command: CommandModifyStake) {
    if (this.mapPeer.has(command.publicKey)) {
      const peer: Peer = this.mapPeer.get(command.publicKey) as Peer;
      if (peer.stake + command.stake < 0) {
        command.stake = -1 * peer.stake;
      }
      this.quorum = this.quorum + command.stake;
      peer.stake = peer.stake + command.stake;
      this.mapPeer.set(command.publicKey, peer);
      await this.updateStateData(Blockchain.STATE_PEER_IDENT + command.publicKey, peer.stake.toString());
    }
  }

  private async setDecision(ns: string, origin: string, data: string) {
    const keyTaken = Blockchain.STATE_DECISION_TAKEN + ns;
    if ((await this.getState(keyTaken)).length) {
      return;
    }

    const key = Blockchain.STATE_DECISION_IDENT + ns;
    try {
      const arrayState = await this.getState(key);
      const mapDecision: Map<string, { stake: number; d: string }> = arrayState.length
        ? new Map(JSON.parse(arrayState[0].value))
        : new Map();
      mapDecision.set(origin, { stake: this.getStake(origin), d: data });
      const stake = [...mapDecision.values()].filter((v) => v.d === data).reduce((p, v) => p + v.stake, 0);
      if (stake >= this.getQuorum()) {
        await this.updateStateData(keyTaken, data);
      } else {
        await this.updateStateData(key, JSON.stringify([...mapDecision]));
      }
    } catch (error: any) {
      Logger.warn(`Blockchain.setDecision() ${key} ${error.toString()}`);
    }
  }

  private async updateBlockData(key: string, data: string) {
    try {
      await this.dbBlockchain.put(key, data);
    } catch (error: any) {
      Logger.warn(`Blockchain.updateBlockData() ${error.toString()}`);
    }
  }

  private async updateStateData(key: string, value: string) {
    try {
      await this.dbState.put(key, value);
    } catch (error: any) {
      Logger.warn(`Blockchain.updateStateData() ${error.toString()}`);
    }
  }
}

==> divachain-0.28.0.tgz: package/src/chain/blockchain.ts <==
/**
 * Copyright (C) 2022 diva.exchange
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Author/Maintainer: Konrad Bächler <konrad@diva.exchange>
 */

import { BlockStruct } from './block';
import fs from 'fs';
import LevelUp from 'levelup';
import LevelDown from 'leveldown';
import path from 'path';
import {
  CommandAddPeer,
  CommandRemovePeer,
  CommandModifyStake,
  CommandData,
  CommandDecision,
  TransactionStruct,
} from './transaction';
import { Server } from '../net/server';
import { Logger } from '../logger';
import { Util } from './util';
import { Readable } from 'stream';

export type Peer = {
  publicKey: string;
  http: string;
  udp: string;
  stake: number;
};

export class Blockchain {
  public static readonly COMMAND_ADD_PEER = 'addPeer';
  public static readonly COMMAND_REMOVE_PEER = 'removePeer';
  public static readonly COMMAND_MODIFY_STAKE = 'modifyStake';
  public static readonly COMMAND_DATA = 'data';
  public static readonly COMMAND_DECISION = 'decision';
  public static readonly STATE_DECISION_IDENT = 'decision:';
  public static readonly STATE_PEER_IDENT = 'peer:';
  public static readonly STATE_DECISION_TAKEN = 'decision:taken:';

  private readonly server: Server;
  private readonly publicKey: string;
  private readonly dbBlockchain: InstanceType<typeof LevelUp>;
  private readonly dbState: InstanceType<typeof LevelUp>;

  private height: number = 0;
  private mapBlocks: Map<number, BlockStruct> = new Map();
  private latestBlock: BlockStruct = {} as BlockStruct;

  private mapPeer: Map<string, Peer> = new Map();

  private quorum: number = 0;

  static async make(server: Server): Promise<Blockchain> {
    const b = new Blockchain(server);
    if (server.config.bootstrap) {
      await b.clear();
    } else {
      await b.init();
    }
    return b;
  }

  private constructor(server: Server) {
    this.server = server;
    this.publicKey = this.server.getWallet().getPublicKey();

    this.dbBlockchain = LevelUp(LevelDown(path.join(this.server.config.path_blockstore, this.publicKey)), {
      createIfMissing: true,
      errorIfExists: false,
      compression: true,
      cacheSize: 2 * 1024 * 1024, // 2 MB
    });

    this.dbState = LevelUp(LevelDown(path.join(this.server.config.path_state, this.publicKey)), {
      createIfMissing: true,
      errorIfExists: false,
      compression: true,
      cacheSize: 2 * 1024 * 1024, // 2 MB
    });
  }

  private async init() {
    this.height = 0;
    this.mapBlocks = new Map();
    this.latestBlock = {} as BlockStruct;
    await this.dbState.clear();

    await new Promise((resolve, reject) => {
      this.dbBlockchain
        .createReadStream()
        .on('data', async (data) => {
          const block: BlockStruct = JSON.parse(data.value) as BlockStruct;
          this.updateCache(block);
          await this.processState(block);
        })
        .on('end', resolve)
        .on('error', reject);
    });
  }

  async shutdown() {
    await this.dbBlockchain.close();
    await this.dbState.close();
  }

  async clear() {
    await this.dbBlockchain.clear();
    await this.dbState.clear();

    this.height = 0;
    this.mapBlocks = new Map();
    this.latestBlock = {} as BlockStruct;
    this.mapPeer = new Map();
  }

  async reset(genesis: BlockStruct) {
    await this.clear();
    await this.updateBlockData(String(1).padStart(16, '0'), JSON.stringify(genesis));
    await this.init();
  }

  add(block: BlockStruct): boolean {
    if (
      this.height + 1 !== block.height ||
      block.previousHash !== this.latestBlock.hash ||
      !this.server.getValidation().validateBlock(block)
    ) {
      return false;
    }

    this.updateCache(block);

    (async (b: BlockStruct) => {
      await this.updateBlockData(String(b.height).padStart(16, '0'), JSON.stringify(b));
      await this.processState(b);
    })(block);

    return true;
  }

  private updateCache(block: BlockStruct) {
    this.height = block.height;
    this.latestBlock = block;

    // cache
    this.mapBlocks.set(this.height, block);
    if (this.mapBlocks.size > this.server.config.blockchain_max_blocks_in_memory) {
      this.mapBlocks.delete(this.height - this.server.config.blockchain_max_blocks_in_memory);
    }
  }

  async getRange(gte: number, lte: number): Promise<Array<BlockStruct>> {
    lte = Math.floor(lte < 1 ? this.height : lte);
    gte = gte <= this.height ? (gte < 1 ? 1 : Math.floor(gte)) : this.height;
    lte = lte <= this.height ? lte : this.height;
    gte = lte - gte > 0 ? gte : lte;
    gte = lte - gte >= this.server.config.api_max_query_size ? lte - this.server.config.api_max_query_size + 1 : gte;

    // cache available?
    if (this.mapBlocks.has(gte) && this.mapBlocks.has(lte)) {
      const start = this.mapBlocks.size - this.height + gte - 1;
      const end = start + lte - gte + 1;
      return [...this.mapBlocks.values()].slice(start, end);
    }

    const a: Array<BlockStruct> = [];
    return new Promise((resolve, reject) => {
      this.dbBlockchain
        .createValueStream({ gte: String(gte).padStart(16, '0'), lte: String(lte).padStart(16, '0') })
        .on('data', (data) => {
          a.push(JSON.parse(data));
        })
        .on('end', () => {
          resolve(a);
        })
        .on('error', reject);
    });
  }

  async getPage(page: number, size: number): Promise<Array<BlockStruct>> {
    page = page < 1 ? 1 : Math.floor(page);
    size =
      size < 1 || size > this.server.config.api_max_query_size
        ? this.server.config.api_max_query_size
        : Math.floor(size);

    let gte = this.height - page * size + 1;
    gte = gte < 1 ? 1 : gte;

    return this.getRange(gte, gte + size - 1);
  }

  async searchBlocks(search: string = ''): Promise<Array<BlockStruct>> {
    return new Promise((resolve, reject) => {
      const a: Array<BlockStruct> = [];
      const stream: Readable = this.dbBlockchain.createValueStream({ reverse: true }) as Readable;
      stream
        .on('data', (data) => {
          data.indexOf(search) > -1 && a.push(JSON.parse(data));
          if (a.length === this.server.config.api_max_query_size) {
            stream.destroy();
            resolve(a.reverse());
          }
        })
        .on('end', () => {
          resolve(a.reverse());
        })
        .on('error', reject);
    });
  }

  async getTransaction(origin: string, ident: string): Promise<{ height: number; transaction: TransactionStruct }> {
    return new Promise((resolve, reject) => {
      // cache
      for (const b of [...this.mapBlocks.values()]) {
        const t = b.tx.find((t: TransactionStruct) => t.origin === origin && t.ident === ident);
        if (t) {
          return resolve({ height: b.height, transaction: t });
        }
      }

      // disk
      let b: BlockStruct = {} as BlockStruct;
      let t: TransactionStruct = {} as TransactionStruct;
      this.dbBlockchain
        .createValueStream()
        .on('data', (data) => {
          if (!t.origin) {
            b = JSON.parse(data) as BlockStruct;
            t =
              b.tx.find((t: TransactionStruct) => t.origin === origin && t.ident === ident) ||
              ({} as TransactionStruct);
          }
        })
        .on('end', () => {
          t.origin ? resolve({ height: b.height, transaction: t }) : reject(new Error('Not Found'));
        })
        .on('error', reject);
    });
  }

  async getState(key: string): Promise<{ key: string; value: string } | false> {
    return new Promise((resolve) => {
      this.dbState.get(key, (error, value: Buffer) => {
        error ? resolve(false) : resolve({ key: key, value: value.toString() });
      });
    });
  }

  async searchState(search: string = ''): Promise<Array<{ key: string; value: any }>> {
    return new Promise((resolve, reject) => {
      const a: Array<{ key: string; value: any }> = [];
      const stream: Readable = this.dbState.createReadStream({ reverse: true }) as Readable;
      stream
        .on('data', (data) => {
          (data.key.toString() + data.value.toString()).indexOf(search) > -1 &&
            a.push({
              key: data.key.toString(),
              value: data.value.toString(),
            });
          if (a.length === this.server.config.api_max_query_size) {
            stream.destroy();
            resolve(a);
          }
        })
        .on('end', () => {
          resolve(a);
        })
        .on('error', reject);
    });
  }

  getLatestBlock(): BlockStruct {
    return this.latestBlock;
  }

  getHeight(): number {
    return this.height;
  }

  getStake(publicKey: string): number {
    try {
      return this.getPeer(publicKey).stake;
    } catch (error) {
      return 0;
    }
  }

  getTotalQuorum(): number {
    if (this.quorum <= 0) {
      throw new Error('Invalid network quorum');
    }

    return this.quorum;
  }

  getQuorum(): number {
    if (this.quorum <= 0) {
      throw new Error('Invalid network quorum');
    }

    return (2 * this.quorum) / 3; // PBFT, PoS
  }

  getMapPeer(): Map<string, Peer> {
    return this.mapPeer;
  }

  hasPeer(publicKey: string): boolean {
    return this.mapPeer.has(publicKey);
  }

  getPeer(publicKey: string): Peer {
    return this.mapPeer.get(publicKey) as Peer;
  }

  hasNetworkHttp(http: string): boolean {
    for (const v of [...this.mapPeer]) {
      if (v[1].http === http) {
        return true;
      }
    }
    return false;
  }

  async getPerformance(height: number): Promise<{ timestamp: number }> {
    let ts: number;
    try {
      ts = Number((await this.dbState.get('debug-performance-' + height)).toString());
    } catch (error) {
      ts = 0;
    }
    return { timestamp: ts };
  }

  static genesis(p: string): BlockStruct {
    if (!fs.existsSync(p)) {
      throw new Error('Genesis Block not found at: ' + p);
    }
    const b: BlockStruct = JSON.parse(fs.readFileSync(p).toString());
    b.hash = Util.hash(b.previousHash + b.version + b.height + JSON.stringify(b.tx));
    return b;
  }

  private async processState(block: BlockStruct) {
    if (this.server.config.debug_performance) {
      await this.updateStateData('debug-performance-' + this.height, new Date().getTime().toString());
    }

    for (const t of block.tx) {
      for (const c of t.commands) {
        switch (c.command) {
          case Blockchain.COMMAND_ADD_PEER:
            await this.addPeer(c as CommandAddPeer);
            break;
          case Blockchain.COMMAND_REMOVE_PEER:
            await this.removePeer(c as CommandRemovePeer);
            break;
          case Blockchain.COMMAND_MODIFY_STAKE:
            await this.modifyStake(c as CommandModifyStake);
            break;
          case Blockchain.COMMAND_DATA:
            await this.updateStateData((c as CommandData).ns + ':' + t.origin, (c as CommandData).d);
            break;
          case Blockchain.COMMAND_DECISION:
            await this.setDecision((c as CommandDecision).ns, t.origin, (c as CommandDecision).d);
            break;
        }
      }
    }
  }

  private async addPeer(command: CommandAddPeer) {
    if (this.mapPeer.has(command.publicKey)) {
      return;
    }

    const peer: Peer = {
      publicKey: command.publicKey,
      http: command.http,
      udp: command.udp,
      stake: 0,
    };
    this.mapPeer.set(command.publicKey, peer);
    await this.updateStateData(Blockchain.STATE_PEER_IDENT + command.publicKey, peer.stake.toString());
  }

  private async removePeer(command: CommandRemovePeer) {
    if (this.mapPeer.has(command.publicKey)) {
      const peer: Peer = this.mapPeer.get(command.publicKey) as Peer;
      this.quorum = this.quorum - peer.stake;

      this.mapPeer.delete(command.publicKey);
      await this.deleteStateData(Blockchain.STATE_PEER_IDENT + command.publicKey);
    }
  }

  private async modifyStake(command: CommandModifyStake) {
    if (this.mapPeer.has(command.publicKey)) {
      const peer: Peer = this.mapPeer.get(command.publicKey) as Peer;
      if (peer.stake + command.stake < 0) {
        command.stake = -1 * peer.stake;
      }
      this.quorum = this.quorum + command.stake;
      peer.stake = peer.stake + command.stake;
      this.mapPeer.set(command.publicKey, peer);
      await this.updateStateData(Blockchain.STATE_PEER_IDENT + command.publicKey, peer.stake.toString());
    }
  }

  private async setDecision(ns: string, origin: string, data: string) {
    const keyTaken = Blockchain.STATE_DECISION_TAKEN + ns;
    if (await this.getState(keyTaken)) {
      return;
    }

    const key = Blockchain.STATE_DECISION_IDENT + ns;
    try {
      const state = await this.getState(key);
      const mapDecision: Map<string, { stake: number; d: string }> = state
        ? new Map(JSON.parse(state.value))
        : new Map();
      mapDecision.set(origin, { stake: this.getStake(origin), d: data });
      const stake = [...mapDecision.values()].filter((v) => v.d === data).reduce((p, v) => p + v.stake, 0);
      if (stake >= this.getQuorum()) {
        await this.updateStateData(keyTaken, JSON.stringify([...mapDecision]));
        await this.deleteStateData(key);
      } else {
        await this.updateStateData(key, JSON.stringify([...mapDecision]));
      }
    } catch (error: any) {
      Logger.warn(`Blockchain.setDecision() ${key} ${error.toString()}`);
    }
  }

  private async updateBlockData(key: string, data: string) {
    try {
      await this.dbBlockchain.put(key, data);
    } catch (error: any) {
      Logger.warn(`Blockchain.updateBlockData() ${error.toString()}`);
    }
  }

  private async updateStateData(key: string, value: string) {
    try {
      await this.dbState.put(key, value);
    } catch (error: any) {
      Logger.warn(`Blockchain.updateStateData() ${error.toString()}`);
    }
  }

  private async deleteStateData(key: string) {
    try {
      await this.dbState.del(key);
    } catch (error: any) {
      Logger.warn(`Blockchain.updateStateData() ${error.toString()}`);
    }
  }
}
